open Printf

	exception INDEXOVERFLOW of string
	exception INDEXMISMATCH of string

	type binaryMatrix = int * int * bool array array ref
	
	let make x y ival =
		let matrix = Array.make_matrix x y ival
		in (x, y, ref matrix)

	let set (xdim, ydim, matrix) x y newval =
		if (x >= 0 && x < xdim && y >=0 && y < ydim) then
			!matrix.(x).(y) <- newval
		else raise (INDEXOVERFLOW ("SET [" ^ (string_of_int x) ^ "," ^ (string_of_int y) ^ "] to a " ^ (string_of_int xdim) ^ "*" ^ (string_of_int ydim) ^ " matrix"))

	let get (xdim, ydim, matrix) x y =
		if (x < xdim && y < ydim) then
			!matrix.(x).(y)
		else 	raise (INDEXOVERFLOW ("GET [" ^ (string_of_int x) ^ "," ^ (string_of_int y) ^ "] from a " ^ (string_of_int xdim) ^ "*" ^ (string_of_int ydim) ^ " matrix"))
	
	let add (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2) =
		let addition (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2) =
			let prodMatrix = Array.make_matrix xdim1 ydim2 false
			in for i = 0 to xdim1 - 1 do
				for j = 0 to ydim1 - 1 do
					prodMatrix.(i).(j) <- !matrix1.(i).(j) || !matrix2.(i).(j)
				done
			   done;
			   (xdim1, ydim2, ref prodMatrix)
		in	if (xdim1 = xdim2 && ydim1 = ydim2) then
				addition (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2)
			else raise (INDEXMISMATCH ("ADD a " ^ (string_of_int xdim1) ^ "*" ^ (string_of_int ydim1) ^ " matrix with a " ^ (string_of_int xdim2) ^ "*" ^ (string_of_int ydim2) ^ " matrix"))

	let mul (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2) =
		let multiply (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2) =
			let prodMatrix = Array.make_matrix xdim1 ydim2 false
			in for i = 0 to xdim1 - 1 do
				for j = 0 to ydim1 - 1 do
					prodMatrix.(i).(j) <- !matrix1.(i).(j) && !matrix2.(i).(j)
				done
			   done;
			   (xdim1, ydim2, ref prodMatrix)
		in	if (xdim1 = xdim2 && ydim1 = ydim2) then
				multiply (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2)
			else raise (INDEXMISMATCH ("MUL a " ^ (string_of_int xdim1) ^ "*" ^ (string_of_int ydim1) ^ " matrix with a " ^ (string_of_int xdim2) ^ "*" ^ (string_of_int ydim2) ^ " matrix"))

	let prod (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2)=
		let product (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2) =
			let prodMatrix = Array.make_matrix xdim1 ydim2 false
			in for i = 0 to xdim1 - 1 do
				for j = 0 to ydim2 - 1 do
					for k = 0 to xdim2 - 1 do
						prodMatrix.(i).(j) <- (prodMatrix.(i).(j) || (!matrix1.(i).(k) && !matrix2.(k).(j)))
					done
				done
			   done;
			   (xdim1, ydim2, ref prodMatrix)
		in	if (ydim1 = xdim2) then
				product (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2)
			else raise (INDEXMISMATCH ("PROD a " ^ (string_of_int xdim1) ^ "*" ^ (string_of_int ydim1) ^ " matrix with a " ^ (string_of_int xdim2) ^ "*" ^ (string_of_int ydim2) ^ " matrix"))

	let prodOr (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2)=
		let product (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2) =
			let prodMatrix = Array.make_matrix xdim1 ydim2 false
			in for i = 0 to xdim1 - 1 do
				for j = 0 to ydim2 - 1 do
					for k = 0 to xdim2 - 1 do
						if (k != i) then
							prodMatrix.(i).(j) <- (prodMatrix.(i).(j) || (!matrix1.(i).(k) && !matrix2.(k).(j)))
					done
				done
			   done;
			   (xdim1, ydim2, ref prodMatrix)
		in	if (ydim1 = xdim2) then
				product (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2)
			else raise (INDEXMISMATCH ("PROD a " ^ (string_of_int xdim1) ^ "*" ^ (string_of_int ydim1) ^ " matrix with a " ^ (string_of_int xdim2) ^ "*" ^ (string_of_int ydim2) ^ " matrix"))

	let prodAnd (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2) =
		let product (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2) =
			let prodMatrix = Array.make_matrix xdim1 ydim2 true
			in for i = 0 to xdim1 - 1 do
				for j = 0 to ydim2 - 1 do
					for k = 0 to xdim2 - 1 do
						if (k != i) then
							prodMatrix.(i).(j) <- (prodMatrix.(i).(j) && (not !matrix1.(i).(k) || !matrix2.(k).(j)))
					done
				done
			   done;
			   (xdim1, ydim2, ref prodMatrix)
		in	if (ydim1 = xdim2) then
				product (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2)
			else raise (INDEXMISMATCH ("PRODAND a " ^ (string_of_int xdim1) ^ "*" ^ (string_of_int ydim1) ^ " matrix with a " ^ (string_of_int xdim2) ^ "*" ^ (string_of_int ydim2) ^ " matrix"))

	let printMatrix (xdim, ydim, matrix) =
		for i = 0 to xdim -1 do
			for j = 0 to ydim - 1 do
				if !matrix.(i).(j) then printf "1 " else printf "0 "
			done;
			printf "\n"
		done;
		printf "\n"

	let isAll (xdim, ydim, matrix) b =
		let i = ref 0
		and ok = ref false
		in 	while (!i < xdim) && not !ok do
			let j = ref 0
			in 	while (!j < ydim) && !matrix.(!i).(!j) = b do
					j := !j + 1
				done;
				if (!j = ydim) then
					i := !i + 1
				else
					ok := true
			done;
			not (!i < xdim)

	let comp (xdim1, ydim1, matrix1) (xdim2, ydim2, matrix2) =
		let	i = ref 0
		and	eq = ref true
		and	xdim = min xdim1 xdim2
		and  ydim = min ydim1 ydim2
		in 	while (!i < xdim) && !eq do
			let j = ref 0 
			in	while (!j < ydim) && !matrix1.(!i).(!j) = !matrix2.(!i).(!j) do
					j := !j + 1
				done;
				if (!j < ydim) then
					eq := false;
				i := !i  + 1
			done;
			!eq
