
// 	type exp = EXP1 of string * exp ref | EXP2 of string * exp ref * exp ref | VAR of string | BVALUE of bool | VALUE of int

#include <stdio.h>
#include <string.h>

#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/callback.h>
#include <caml/alloc.h>

#include "trans2.h"

int optype(char *op)
{
	int typeID;
	
	switch (*op) {
		case '+':
		case '-':
		case '*':
		case '/':
		case '%':
			typeID = 102;
			break;
		case '<': // "<=":
		case '>': // ">=":
		case '=':
			typeID = 103;
			break;
		case '&': // "&&":
		case '|': // "||":
		case '!':
			typeID = 104;
			break;
	}
	return typeID;
}
				
value mkExp1ConNode(value var, value ex)
{
	union Valuetag val; 
	
	CAMLparam2 (var, ex);
	strcpy(val.svalue, String_val(var));
	CAMLreturn ((value)(NewConNode(val, optype(String_val(var)),NULL, (ConNode *)ex)));
}

value mkExp2ConNode(value var, value ex1, value ex2)
{
	union Valuetag val; 
	char *op = NULL;
	
	CAMLparam3 (var, ex1, ex2);
	op = String_val(var);
	switch (*op) {
		case '=' : strcpy(val.svalue, "=="); break;
		default : strcpy(val.svalue, String_val(var));
	}
	CAMLreturn ((value)(NewConNode(val, optype(String_val(var)), (ConNode *)ex2, (ConNode *)ex1)));
}

value mkVarConNode(value var)
{
	union Valuetag val; 
	
	CAMLparam1 (var);
	strcpy(val.svalue, String_val(var));
	CAMLreturn ((value)(NewConNode(val, 100, NULL, NULL))); // numeric var
}

value mkBoolConNode(value var)
{
	union Valuetag val; 	
	
	CAMLparam1 (var);
	val.bvalue = Bool_val(var);
	CAMLreturn ((value)(NewConNode(val, 105, NULL, NULL))); // constant
}

value mkIntConNode(value var)
{
	union Valuetag val; 
	
	CAMLparam1 (var);
	val.fvalue = Int_val(var);
	CAMLreturn ((value)(NewConNode(val, 105, NULL, NULL))); // constant
}

value freeConNode(value node)
{
	ConNode *conNode;

	CAMLparam1(node);
	conNode = (ConNode *)node;
	if (conNode -> left != NULL)
		freeConNode((value)(conNode -> left));
	if (conNode -> right != NULL)
		freeConNode((value)(conNode -> right));
	free(conNode);
	CAMLreturn (Val_unit);
}

void printConNode(ConNode *p)
{
	if (!p) return;
	
	switch(p -> NodeType) {
		case 100: printf("%s", p -> NodeValue.svalue); break;
		case 102:
		case 103:
		case 104:
			printConNode(p -> left);
			printf("%s", p -> NodeValue.svalue);
			printConNode(p -> right);
			break;
		case 105: printf("%d", (int)p -> NodeValue.fvalue); break;
	};
}

value solver(value cn, value num_of_var, value vnl, value vvl)
{
	Constraint *cons=NULL;
	ConSet *cset = NULL;
	REAL *varValue = NULL;
	char **varName = NULL;
	int i, res;
	
	CAMLparam2 (cn, num_of_var);
//	printf("------------------------------------------------------------------\n");
//	printConNode((ConNode *)cn);
//	printf("\n");				
	cons = NewConstraint((ConNode *)cn, 0);
	cset = NewConSet(cons, NULL);
    	res = bonus(cset, NULL, &varName, &varValue);
	if (res == 1) {
		for(i = 0;i < Int_val(num_of_var);i++) {
//			printf("%s = %d", varName[i], (int)varValue[i]);
			callback2(*caml_named_value("getVarNameList"), copy_string(varName[i]), vnl);
			callback2(*caml_named_value("getVarValueList"), Val_int(varValue[i]), vvl);
		}
//		printf("\n");						
	}
	freeConNode(cn);
	free(cons);
	free(cset);
	CAMLreturn (Val_int(res));
}

value lmalloc(value size)
{
	CAMLparam1 (size);
	CAMLreturn ((value)(malloc(Int_val(size))));
}

value lfree(value p)
{
	CAMLparam1 (p);
	printf("To free a node at %0x:%d\n", (int)p, ((ConNode *)p)->NodeType);
	free((ConNode *)p);
	CAMLreturn (Val_unit);
}
