open Exp2Node

	exception INVALIDSUSTITUTION
	
	type exp = EXP1 of string * exp ref | EXP2 of string * exp ref * exp ref | VAR of string | BVALUE of bool | VALUE of int
	type conNodeRef = Exp2Node.conNodeRef

	let rec exp2String ex =
	match !ex with
		|EXP1(opn, ex1) -> opn^(exp2String ex1)
		|EXP2(opn, ex1, ex2) -> (exp2String ex1)^" "^opn^" "^(exp2String ex2)
		|VAR(v) -> v
		|BVALUE(b) -> if b then "" else "_false"
		|VALUE(v) -> string_of_int v

	let expList2String el = 
		let mergeExp e es =
			let pre = exp2String(e)
			in if es = "" then pre else pre^", "^es
		in List.fold_right mergeExp !el ""

	let rec constantEval  ex = match !ex with
		|EXP1(opn, ex1) ->
			let ex1' = constantEval ex1
			in begin
				match (opn, !ex1') with
					|("!", BVALUE(b)) -> ref (BVALUE(not b))
					|_ -> ref (EXP1(opn, ex1'))
			end
		|EXP2(opn, ex1, ex2) ->
			let ex1' = constantEval ex1
			and ex2' = constantEval ex2
			in begin
				match (opn, !ex1', !ex2') with
					|("&&", BVALUE(false), _) -> ex1'
					|("&&", _, BVALUE(false)) -> ex2'
					|("&&", BVALUE(true), _) -> ex2'
					|("&&", _, BVALUE(true)) -> ex1'		
					|("||", BVALUE(true), _) -> ex1'
					|("||", _, BVALUE(true)) -> ex2'
					|("||", BVALUE(false), _) -> ex2'
					|("||", _, BVALUE(false)) -> ex1'								
					|("+", VALUE(v1), VALUE(v2)) -> ref (VALUE(v1+v2))
					|("-", VALUE(v1), VALUE(v2)) -> ref (VALUE(v1-v2))
					|("*", VALUE(v1), VALUE(v2)) -> ref (VALUE(v1*v2))
					|("/", VALUE(v1), VALUE(v2)) -> ref (VALUE(v1/v2))
					|("%", VALUE(v1), VALUE(v2)) -> ref (VALUE(v1 mod v2))
					|("<", VALUE(v1), VALUE(v2)) -> ref (BVALUE(v1<v2))
					|("<=", VALUE(v1), VALUE(v2)) -> ref (BVALUE(v1<=v2))
					|(">", VALUE(v1), VALUE(v2)) -> ref (BVALUE(v1>v2))						
					|(">=", VALUE(v1), VALUE(v2)) -> ref (BVALUE(v1>=v2))
					|("=", VALUE(v1), VALUE(v2)) -> ref (BVALUE(v1=v2))						
					|("!=", VALUE(v1), VALUE(v2)) -> ref (BVALUE(v1!=v2))
					|_ -> ref (EXP2(opn, ex1', ex2'))
			end
		|_ -> ex

	let substExp vl el ex = 
		let rec pairing (ll, rl) = match (ll, rl) with
		| ([], _) -> []
		| (_, []) -> []
		|(le::lt, re::rt) -> (le, re)::pairing(lt, rt) in
		let pairs = pairing (!vl,!el) in
		let rec subst e = match !e with
			|EXP1(op, ex1) -> ref (EXP1(op, subst ex1))
			|EXP2(op, ex1, ex2) -> ref (EXP2(op, subst ex1, subst ex2))
			|VAR(v) ->
				let rec substVar subs= match subs with
					| [] -> e
					| (le, re)::rest ->
						match !le with
						|VAR(v') -> if (compare v v') = 0 then re else substVar rest
						|_ -> raise INVALIDSUSTITUTION
				in substVar pairs
			|_ -> e
		in constantEval (subst ex)

	let isBoolConstant ex b = match !ex with
		|BVALUE(bv) -> b = bv
		|_ -> false

	let isVar ex = match !ex with
		|VAR(_) -> true
		|_ -> false

	let isVal ex = match !ex with
		|VALUE(_) -> true
		|BVALUE(_) -> true
		|_ -> false
		
	let conjunctBoolExp ex1 ex2 =
(*		if (isBoolConstant ex1 true) || (isBoolConstant ex2 false) then
			constantEval ex2
		else if (isBoolConstant ex2 true)  || (isBoolConstant ex1 false)then
			constantEval ex1
		else *)constantEval (ref (EXP2("&&", ex1, ex2)))

	let numofVar ex =
		let vall = ref [] in
		let rec getVar e =
			match !e with
			|EXP1(opn, ex1) -> getVar ex1
			|EXP2(opn, ex1, ex2) -> getVar ex1; getVar ex2
			|VAR(v) -> if not (List.mem v !vall) then vall := v :: !vall
			|_ -> ()
		in getVar ex; List.length !vall

