open Expression
open Graph

	type exp = Expression.exp
	type action = Graph.action
	type constrained_action = action ref * exp ref
	type formula = BVALUE of bool
				|DIAMOND of constrained_action ref * formula ref
				|BOX of constrained_action ref * formula ref
				|AU of formula ref * formula ref
				|EU of formula ref * formula ref
				|AG of formula ref
				|EG of formula ref
				|NOT of formula ref
				|AND of formula ref * formula ref
				|OR of formula ref * formula ref
	
	let rec formula2String f = 
		let atomicFormula2String (ac, ex) = (Graph.action2String ac)^"["^(Expression.exp2String ex)^"]"
		in match !f with
		|BVALUE(b) -> if b then "_true" else "_false"
		|BOX(a, xf) -> "{"^(atomicFormula2String !a)^"} "^(formula2String xf)
		|DIAMOND(a, xf) -> "<"^(atomicFormula2String !a)^"> "^(formula2String xf)
		|AU(f1,f2) -> "_A["^(formula2String f1)^" _U "^(formula2String f2)^"]"
		|EU(f1,f2) -> "_E["^(formula2String f1)^" _U "^(formula2String f2)^"]"
		|AG(gf) -> "_A _G["^(formula2String gf)^"]"
		|EG(gf) -> "_E _G["^(formula2String gf)^"]"
		|NOT(nf) -> "(_not "^formula2String nf^")"
		|AND(f1,f2) -> "("^(formula2String f1)^" _and "^(formula2String f2)^")"
		|OR(f1,f2) -> "("^(formula2String f1)^" _or "^(formula2String f2)^")"
		