open Expression
open Printf

	exception INVALIDSTATE of string
	
	type exp = Expression.exp
	type assign = NONE | ASSIGN of exp ref list ref * exp ref list ref
	type action = INPUT of string * exp ref list ref | OUTPUT of string * exp ref list ref | TAU
				| DELTA of int (* 0 - to a quiescence; 1 - is a quiescence *)
	type ntype = INIT | PROGRESS | ACCEPTANCE
	type node = string * string * exp ref list ref * ntype
	type edge = int * exp ref * assign ref * action ref * int

	let compareNode (nid1, _, _, _) (nid2, _, _, _) = compare nid1 nid2
	let compareEdge (s1, _, _, _, t1) (s2, _, _, _, t2) =
		let res = compare s1 s2
		in if res != 0 then res else compare t1 t2

	let is_act_eq alpha beta =
	match !alpha with
		|INPUT(a, _) -> begin
			match !beta with
				|INPUT(b, _) -> a = b
				|_ -> false
			end
		|OUTPUT(a, _) -> begin
			match !beta with
				|OUTPUT(b, _) -> a = b
				|_ -> false
			end
		|TAU -> begin
			match !beta with
				|TAU -> true
				|_ -> false
			end
		|DELTA(v) -> begin
			match !beta with
				|DELTA(v') -> v = v'
				|_ -> false
			end			

	let action2Assign a seqno = match !a with
		|INPUT(_, varl) ->
			let seqno_str = string_of_int seqno in
			let symVar var = match !var with
				|VAR(v) -> ref (VAR("_"^seqno_str^v))
				|_ -> var in
			let nvarl = List.map symVar !varl
			in ref ((ASSIGN(varl, ref nvarl)))
		|_ -> ref NONE

	let symVal el =
		let initVal ex = match !ex with
			|VAR(v) -> ref (VAR("_0"^v))
			|_ -> raise (INVALIDSTATE("Not a VAR to be symbolized:"^(exp2String ex)))
		in ref (List.map initVal !el)

	let rec action2String a =
	match !a with
		|INPUT(c, exl) -> c^"?"^(Expression.expList2String exl)
		|OUTPUT(c, exl) -> c^"!"^(Expression.expList2String exl)
		|TAU -> "_tau"
		|DELTA(v) -> "_delta"

	let assign2String a =
	match !a with
		|NONE -> ""
		|ASSIGN(el, er) -> "["^(Expression.expList2String el)^" := "^(Expression.expList2String er)^"]"

	let mergeAssign assign1 assign2 = 
		let merged =
		match (!assign1, !assign2) with
		|(NONE, _) -> assign2
		|(_, NONE) -> assign1
		|(ASSIGN(vl1, el1), ASSIGN(vl2, el2)) -> 
			let rec dumpCheck vl el nvl = match (!vl, !el) with
				|([], _) -> (ref [], ref [])
				|(_, []) -> (ref [], ref [])
				|(lv::lt, re::rt) -> 
					if List.mem lv !nvl then 
						dumpCheck (ref lt) (ref rt) nvl 
					else let (vll, ell) = dumpCheck (ref lt) (ref rt) nvl
						in (ref (lv::!vll), ref (re::!ell)) in
			let (vl1', el1') = dumpCheck vl1 el1 vl2
			and el2' = List.map (Expression.substExp vl1 el1) !el2 
			in ref (ASSIGN(ref (List.append !vl2 !vl1'), ref (List.append el2' !el1')))
		in (*printf "Assign1:%s.\n" (assign2String assign1);
			printf "Assign2:%s.\n" (assign2String assign2);
			printf "Assign:%s.\n" (assign2String merged); *)
			merged

	let substCondition b a =
		match !a with
		|NONE -> b
		|ASSIGN(vl, el) -> substExp vl el b
	
