open Printf
open STGAParser
open BinaryMatrix
open Expression
open NodeSet
open EdgeSet
open SymbolicGraph
open Slicing
open TCGen
open Options

let _ =
  let _ = Callback.register "getVarNameList" getVarNameList
  and _ = Callback.register "getVarValueList" getVarValueList 
  and fname = ref 0
  and startTime = ref 0.0
  and endTime = ref 0.0
  and errorLexMsg lexbuf info =
    let pos = Lexing.lexeme_start_p lexbuf
    in printf "%s at Line %d, Offset %d, Position %d(%d)\n" info pos.Lexing.pos_lnum pos.Lexing.pos_bol pos.Lexing.pos_cnum !STGALexer.lineNo in
  let main f =
    let fp = open_in f in
    let lexbuf = Lexing.from_channel fp in try
    let (ns, es, mib, fl)  = STGAParser.main STGALexer.token lexbuf
    and i = ref 0 in
    let testWithConstraint seqno cf =
    	let (cv, ns_slice, es_slice) = Slicing.slicing (ns, es) cf
    	and tcNum = ref 0
	in	printf "# Slice w.r.t constraint no.%d: %s \n" !seqno (Formula.formula2String cf);
		SymbolicGraph.reachabilityCheck (ns_slice, es_slice);
		(*SymbolicGraph.printGraph (ns_slice, es_slice);*) 
	    	if not !(Options.slicingOnly) then begin
			startTime := Sys.time();
			tcNum := TCGen.dfsExec (ns_slice, es_slice);
			endTime := Sys.time();
			printf "# %d test sequences created in %Fs.\n" !tcNum (!endTime -. !startTime)
		end;
		seqno := !seqno + 1   
    in
    	Options.logFile := open_out ("."^f^".log");
    	SymbolicGraph.graphInfo := !mib;
    	printf "# A STGA model is ready with %d node(s) and %d edge(s).\n" (!mib).numof (EdgeSet.cardinal !es);
    	printf "# %d constraint(s) is/are concerned.\n" (List.length !fl);
    	SymbolicGraph.reachabilityCheck (ns, es);
	List.iter (testWithConstraint i) !fl;
    	close_in fp;
    	close_out !(Options.logFile)
  with Failure(info) -> errorLexMsg lexbuf info
    | Parsing.Parse_error -> errorLexMsg lexbuf "syntax error"
    | BinaryMatrix.INDEXMISMATCH(info) -> printf "internal error:%s.\n" info
in	for i = 1 to (Array.length Sys.argv) -1 do
		if Sys.argv.(i).[0] != '-' then
			fname := i
		else if Sys.argv.(i) = "-sol" then Options.constraintSolvingEnabled := true
		else if Sys.argv.(i) = "-false" then Options.falseConditionDisabled := false
		else if Sys.argv.(i) = "-slice" then Options.slicingOnly := true
		done;
	if !fname >0 then
		main Sys.argv.(!fname)
	else begin
		printf "Usage: %s [options] <file>\n\n" Sys.argv.(0);
		printf "file: a model file in STGA with constraint definitions\n";
		printf "options:\n";
		printf "\t-sol\tenable constraint solving\n";
		printf "\t-false\tenable false conditions\n";
		printf "\t-slice\tslice the system model only\n"
	end;	
	exit 0

(*
open Exp2Node

let _=
	let p = mkIntConNode(12);
	in free(p);
		exit(0)
*)

