%{
  open Expression
  open NodeSet
  open EdgeSet
  open NodeSet2
  open NodeSetSet
  open Graph
  open SymbolicGraph
  open Formula

  exception FOUND of int
  exception INVALIDNODEREFERENCE
  let nodeSet = ref NodeSet.empty
  let nodeID = ref 0
  let nodeSetID = ref (-1)
  let node_list = ref []
  let makeEdgeSet el =
  	let n = List.length el
  	and es = ref EdgeSet.empty
  	in	begin
  		for i = 0 to n - 1 do
  			es := EdgeSet.add i !es
  	   	done;
  	   	es
  	   	end
  let getNodeID id =
	let nodeDefined (seqno, _) =
  		let (nid, _, _, _) = List.nth !node_list seqno
  		in if id = int_of_string nid then raise (FOUND(seqno))
  	in try
  		NodeSet.iter nodeDefined !nodeSet;
  		-1
  	with FOUND(seqno) -> seqno
%}
%token <int> NUM
%token TRUE FALSE TAU NOT
%token <string> NAME
%token U G
%token A E
%token AND OR LE LEQ GE GEQ EQ NEQ
%token PLUS MINUS MUL DIV MOD
%token ASSIGN
%token EOF
%token SEMICOLON COLON COMMA LPARENTHESE RPARENTHESE LBRACKET RBRACKET LBBRACKET RBBRACKET QMARK EXCLAMATION

%left AND OR
%right NOT
%left LE LEQ GE GEQ EQ NEQ
%left PLUS MINUS
%left MUL DIV MOD

%start main
%type <SymbolicGraph.allNodeSet ref * SymbolicGraph.allEdgeSet ref * SymbolicGraph.graphMIB ref * Formula.formula ref list ref> main

%%
main: node_section SEMICOLON edge_section SEMICOLON init_section equiv_declaration SEMICOLON formula_section
	{	let ea = ref $3
		and edgeSet = makeEdgeSet $3 in
	  	let cm = SymbolicGraph.buildCM nodeSet edgeSet !nodeID ea
	  	in (	nodeSet,
	  		edgeSet, 
	  		ref {	numof=(!nodeID);
	  				initVal=ref $5;
	  				nodeList=$1; 
	  				edgeList=ea;
	  				equivNodeSets=ref $6;
	  				cmG=cm},
	  		ref $8)
	 };

node_section: { ref [] }
	| node node_section
	{	let (_, _, _, ntype) = $1 
		in	nodeSet := NodeSet.add (!nodeID, ntype) !nodeSet;
			nodeID := !nodeID + 1;
			node_list := $1 :: !$2;
			node_list
	};

node: NUM LBBRACKET exp_list RBBRACKET NAME{ if $1 = 0 then ("0", $5, ref $3, INIT) else (string_of_int $1, $5, ref $3, PROGRESS) };

edge_section: { [] }
	| edge edge_section { $1 :: $2};

edge: NUM LPARENTHESE exp_or_null COMMA assign COMMA action RPARENTHESE NUM
	{ 	let seqno1 = getNodeID $1
		and seqno2 = getNodeID $9
		in 	if seqno1 = -1 || seqno2 = -1 then
				raise (SymbolicGraph.NONSENSEGRAPH ("Parsing: " ^ (string_of_int $1) ^ "->" ^ (string_of_int $9) ^ " for undefined nodes."))
			else
				(seqno1, constantEval (ref $3), ref $5, ref $7, seqno2)
	}

init_section : exp_list { $1 }

equiv_declaration : { NodeSetSet.empty }
	| equiv_set equiv_declaration { nodeSetID := !nodeSetID + 1; NodeSetSet.add (!nodeSetID, $1) $2 }

equiv_set : LBBRACKET exp_list RBBRACKET
	{	let ns = ref NodeSet2.empty in
		let makeEquivNodeSet nid = match !nid with
			|VALUE(v) -> ns := NodeSet2.add (getNodeID v) !ns
			|_ -> raise INVALIDNODEREFERENCE
		in 	List.iter makeEquivNodeSet $2;
			ns
	}

exp_or_null: { Expression.BVALUE(true) }
	| exp { $1 };

exp: TRUE { Expression.BVALUE(true) }
	| FALSE { Expression.BVALUE(false) }
	| NUM { Expression.VALUE($1)}
	| NAME { Expression.VAR($1)}
	| exp LE exp { Expression.EXP2("<", ref $1, ref $3) }
	| exp LEQ exp { Expression.EXP2("<=", ref $1, ref $3) }
	| exp GE exp { Expression.EXP2(">", ref $1, ref $3) }
	| exp GEQ exp { Expression.EXP2(">=", ref $1, ref $3) }
	| exp EQ exp { Expression.EXP2("=", ref $1, ref $3) }
	| exp NEQ exp { Expression.EXP2("!=", ref $1, ref $3) }
	| exp AND exp { Expression.EXP2("&&", ref $1, ref $3) }
	| exp OR exp { Expression.EXP2("||", ref $1, ref $3) }
	| NOT exp { Expression.EXP1("!", ref $2) }
	| exp PLUS exp { Expression.EXP2("+", ref $1, ref $3) }
	| exp MINUS exp { Expression.EXP2("-", ref $1, ref $3) }
	| exp MUL exp { Expression.EXP2("*", ref $1, ref $3) }
	| exp DIV exp { Expression.EXP2("/", ref $1, ref $3) }
	| exp MOD exp { Expression.EXP2("%", ref $1, ref $3) }
	| LPARENTHESE exp RPARENTHESE { $2};

assign: { NONE }
	| LBRACKET RBRACKET { NONE }
	| LBRACKET exp_list ASSIGN exp_list RBRACKET { ASSIGN(ref $2, ref $4) };

exp_list: { [] }
	| non_empty_exp_list { List.map Expression.constantEval $1 };

non_empty_exp_list: exp { [ ref $1 ] }
	| exp COMMA exp_list { (ref $1) :: $3 };

action: TAU { TAU }
	| NAME QMARK exp_list { INPUT($1, ref $3) }
	| NAME EXCLAMATION exp_list { OUTPUT($1, ref $3) };

formula_section: { [] }
	| non_empty_formula_list { $1 };

non_empty_formula_list : formula { [ ref $1 ] }
	| formula COMMA non_empty_formula_list { (ref $1) :: $3 };

constrained_action: action LBRACKET exp_or_null RBRACKET { (ref $1, ref $3) };
 
formula: TRUE { BVALUE(true) }
	| FALSE { BVALUE(false) }
	| formula AND formula { AND(ref $1, ref $3) }
	| formula OR formula { OR(ref $1, ref $3) }
	| LBBRACKET constrained_action RBBRACKET formula  { BOX(ref $2, ref $4) }
	| LE constrained_action GE formula { DIAMOND(ref $2, ref $4) }
	| A LBRACKET formula U formula RBRACKET { AU(ref $3, ref $5) }
	| E LBRACKET formula U formula RBRACKET { EU(ref $3, ref $5) }
	| A G LBRACKET formula RBRACKET { AG(ref $4) }
	| E G LBRACKET formula RBRACKET { EG(ref $4) }
	| LPARENTHESE formula RPARENTHESE { $2 };
