open BinaryMatrix
open Graph
open NodeSet
open EdgeSet
open BinaryMatrixSet
open SymbolicGraph
open Formula
open Printf

	exception NONSENSECONSTRAINT
	exception ILLEGALNEGATIVE
	
       let mergeNodeSet ns1 ns2 =
    		let nsDiff = NodeSet.union (NodeSet.diff !ns1 !ns2) (NodeSet.diff !ns2 !ns1)
    		and	ns = ref (NodeSet.inter !ns1 !ns2)
    		and isAcceptance id (nid, ntype) = id = nid && ntype = ACCEPTANCE in
    		let setAcceptance (id, ntype) =
    			if ntype != ACCEPTANCE && ((NodeSet.exists (isAcceptance id) !ns1) || (NodeSet.exists (isAcceptance id) !ns2)) then begin
    				ns := NodeSet.remove (id, ntype) !ns;
    				ns := NodeSet.add (id, ACCEPTANCE) !ns;
    			end
    		in 	NodeSet.iter setAcceptance !ns;
    			NodeSet.union nsDiff !ns

	let setAllAccepted ns =
    		let setAcceptance (id, ntype) =
    			if ntype != ACCEPTANCE then begin
    				ns := NodeSet.remove (id, ntype) !ns;
    				ns := NodeSet.add (id, ACCEPTANCE) !ns;
    			end
    		in 	NodeSet.iter setAcceptance !ns

	let slicing (ns, es) f =
		let nsrc = (!graphInfo).numof
		and na = (!graphInfo).nodeList
		and ea = (!graphInfo).edgeList
		and cmG = (!graphInfo).cmG in
		let filterNode id (nid, _) = nid = id in
		let filterEdge scv tcv nodeSet edgeSet actionAlpha eid  =
			let (sid, _, _, beta, tid) = List.nth !ea eid
			and setAcceptance tid (id, ntype) =
				if id = tid && ntype != ACCEPTANCE then begin
					nodeSet := NodeSet.remove (id, ntype) !nodeSet;
					nodeSet := NodeSet.add (id, ACCEPTANCE) !nodeSet
				end
			in	if (BinaryMatrix.get scv sid 0) && Graph.is_act_eq actionAlpha beta && (BinaryMatrix.get tcv tid 0) then begin
					nodeSet := mergeNodeSet (ref (NodeSet.filter (filterNode sid) !ns)) nodeSet;
					nodeSet := mergeNodeSet (ref (NodeSet.filter (filterNode tid) !ns)) nodeSet;
					NodeSet.iter (setAcceptance tid) !nodeSet;
					edgeSet := EdgeSet.add eid !edgeSet
				end
		and filterEdgeAll scv tcv nodeSet edgeSet eid  =
			let (sid, _, _, _, tid) = List.nth !ea eid
			in 	if (BinaryMatrix.get scv sid 0) && (BinaryMatrix.get tcv tid 0) then begin
					nodeSet := mergeNodeSet (ref (NodeSet.filter (filterNode sid) !ns)) nodeSet;
					nodeSet := mergeNodeSet (ref (NodeSet.filter (filterNode tid) !ns)) nodeSet;
					edgeSet := EdgeSet.add eid !edgeSet
				end in
		let rec lfp cVector nodeSet edgeSet f =
			let cVNew = f cVector in
			let cVector' = BinaryMatrix.add cVector cVNew
		  	in 	(* printMatrix cVector';
		  		printf "-------------->>>\n";
		  		printMatrix cVector; *)
		  		EdgeSet.iter (filterEdgeAll cVNew cVector nodeSet edgeSet) !es;
		    		if (BinaryMatrix.comp cVector' cVector) then (cVector, nodeSet, edgeSet)
		    		else lfp cVector' nodeSet edgeSet f in
              let rec gfp cVector nodeSet edgeSet f =
			let cVNew = f cVector in
			let cVector' = BinaryMatrix.mul cVector cVNew
		  	in 	(* printMatrix cVector';
		  		printf "-------------->>>\n";
		  		printMatrix cVector; *)
		  		EdgeSet.iter (filterEdgeAll cVNew cVector nodeSet edgeSet) !es;
		    		if (BinaryMatrix.comp cVector' cVector) then (cVector, nodeSet, edgeSet)
		    		else gfp cVector' nodeSet edgeSet f in
		let rec slicingGraph f =
			let ns_slice = ref NodeSet.empty
			and es_slice = ref EdgeSet.empty
			in match !f with
				|BVALUE(bv) ->
					let cv = BinaryMatrix.make nsrc 1 bv
					in 	if bv then begin
							(*setAllAccepted ns;  only for symbolic execution of the model with trivial case _true *)
							(cv, ns, es)
						end
						else (cv, ref NodeSet.empty,  ref EdgeSet.empty)
				|AND(f1, f2) ->
				   	let (cv1, ns1, es1) = slicingGraph f1
				   	and (cv2, ns2, es2) = slicingGraph f2 in
				   	let cv = BinaryMatrix.mul cv1 cv2
					in (cv, ref (mergeNodeSet ns1 ns2), ref (EdgeSet.union !es1 !es2))
				|OR(f1, f2) -> 
				   	let (cv1, ns1, es1) = slicingGraph f1
				   	and (cv2, ns2, es2) = slicingGraph f2 in
				   	let cv = BinaryMatrix.add cv1 cv2
					in (cv, ref (mergeNodeSet ns1 ns2), ref (EdgeSet.union !es1 !es2))
				|DIAMOND(action, xf) ->
					let (alpha, _) = !action
					and (cv, ns_slice, es_slice) = slicingGraph xf in
					let cv' = BinaryMatrix.prodOr (SymbolicGraph.buildActionCM (ns, es) alpha) cv
					in 	EdgeSet.iter (filterEdge cv' cv ns_slice es_slice alpha) !es;
						(cv', ns_slice, es_slice)
				|BOX(action, xf) ->
					let (alpha, _) = !action
					and (cv, ns_slice, es_slice) = slicingGraph xf in
					let cv' = BinaryMatrix.prodAnd (SymbolicGraph.buildActionCM (ns, es) alpha) cv
					in 	EdgeSet.iter (filterEdge cv' cv ns_slice es_slice alpha) !es;
						(cv', ns_slice, es_slice)
				|AG(gf) ->
					let (cv, ns_slice, es_slice) = slicingGraph gf in
                                   let ag cV = BinaryMatrix.mul cv (BinaryMatrix.prodAnd cmG cV)
					in gfp cv ns_slice es_slice ag				
				|EG(gf) ->
					let (cv, ns_slice, es_slice) = slicingGraph gf in
                                   let eg cV = BinaryMatrix.mul cv (BinaryMatrix.prodOr cmG cV)
					in gfp cv ns_slice es_slice eg		
				|AU(f1, f2) ->
				   let (cv1, ns1, es1) = slicingGraph f1 
				   and (cv2, ns2, es2) = slicingGraph f2
				   and au cV1 cV2 cV = BinaryMatrix.add cV2 (BinaryMatrix.mul cV1 (BinaryMatrix.prodAnd cmG cV))
				   in lfp cv2 (ref (mergeNodeSet ns1 ns2)) (ref (EdgeSet.union !es1 !es2)) (au cv1 cv2)
				|EU(f1, f2) ->
				   let (cv1, ns1, es1) = slicingGraph f1
				   and (cv2, ns2, es2) = slicingGraph f2
				   and eu cV1 cV2 cV = BinaryMatrix.add cV2 (BinaryMatrix.mul cV1 (BinaryMatrix.prodOr cmG cV))
				   in lfp cv2 (ref (mergeNodeSet ns1 ns2)) (ref (EdgeSet.union !es1 !es2)) (eu cv1 cv2)
				|_ -> raise NONSENSECONSTRAINT
		in slicingGraph f

