open Expression
open ExpNode
open Graph
open NodeSet
open EdgeSet
open StateSet
open SymbolicGraph
open Printf

       type symbolicGraph = SymbolicGraph.symbolicGraph

      	exception VISITEDSTATE

	let isVisited s vs =
		let statecmp s1 s2 = 
			if (compareState s1 s2) = 0 then
				raise VISITEDSTATE
		in try
			StateSet.iter (statecmp s) !vs;
			false
		with VISITEDSTATE -> true

    	let dfsExec (ns, es) =
		let tcNum = ref 0 
		and nsrc = (!graphInfo).numof
		and na = (!graphInfo).nodeList
		and ea = (!graphInfo).edgeList in
		let (initState, initAssign) =
			let (_, _, args, _) = List.nth !na 0
			in 	if (List.length !args) = 0 then
					((0, ref []), NONE)
				else	let initSym = symVal args
					in ((0, (!graphInfo).initVal), ASSIGN(args, (!graphInfo).initVal)) in
		let rec dfs nid path con assign vss solvn solvv =
			EdgeSet.iter (iterStep nid path con assign vss solvn solvv) !es
		and iterStep nid path con lastAssign vss solvn solvv eid =
			let (sid, b, assign, alpha, tid) = List.nth !ea eid
			and isAcceptance aID (nid, ntype) = aID = nid && ntype = ACCEPTANCE
			in if sid = nid (* && not (List.mem eid path) *) then
				let b' = substCondition b lastAssign in
				let bSol = substExp solvn solvv b'
				in if not (!Options.falseConditionDisabled) || not (isBoolConstant bSol false) then
					let con' = conjunctBoolExp con b' in
					let conSol = substExp solvn solvv con'
					in if not (!Options.falseConditionDisabled) || not (isBoolConstant conSol false) then
						let path' = List.append path [eid]
						and assignAlpha = action2Assign alpha ((List.length path)+1) in
						let eval = mergeAssign lastAssign (mergeAssign assign assignAlpha) in
						let nextState = getNextState tid eval
						in begin
							if NodeSet.exists (isAcceptance tid) !ns then begin
								printf "A test sequence (%s) is ready with a constraint %s.\n" (edgeList2String path' na ea) (exp2String con');
								tcNum := !tcNum + 1;
								if !Options.constraintSolvingEnabled && not (isBoolConstant conSol true) then
									let conNode = mkConNode conSol
									and vnl = ref []
									and vvl = ref [] in
									let res = Exp2Node.solver conNode (numofVar conSol) vnl vvl
									in if (res = 1) then begin
										printf "A solution is here:\n";								
										printf "%s := %s.\n" (expList2String vnl) (expList2String vvl)
									end
							end;	
(*							if not (StateSet.mem nextState !vss) then begin *)
							if not (isVisited nextState vss) then begin
								logState nextState;							
								vss := StateSet.add nextState !vss;
								if not (List.mem eid path) || (isBoolConstant bSol true) then begin
									dfs tid path' con' eval vss solvn solvv;
								end
								else if (!Options.constraintSolvingEnabled) then
									let conNode = mkConNode conSol
									and vnl = ref []
									and vvl = ref [] in
									let res = Exp2Node.solver conNode (numofVar conSol) vnl vvl
									in begin
										printf "A DEFERRED loop is detected (%s) with the last condition %s.\n" (edgeList2String path' na ea) (exp2String b');
										if (res = 1) then
											dfs tid path' con' eval vss (ref (List.append !solvn !vnl)) (ref (List.append !solvv !vvl))						
									end
							end
							else	fprintf !(Options.logFile) "-----------------------------\n";
								(* StateSet.iter logState !vss; *)
								fprintf !(Options.logFile) "The number of visited states: %d.\n" (StateSet.cardinal !vss);
								fprintf  !(Options.logFile) ">"; 
								logState nextState;
								fprintf  !(Options.logFile) "-----------------------------\n"
						end
			in 	dfs 0 [] (ref (BVALUE(true))) (ref initAssign) (ref (StateSet.singleton initState)) (ref []) (ref []);
				!tcNum

